#ifndef SourceFileH
#define SourceFileH

#include <vector>
#include <map>

#include "ResizeCfg.h"

struct SourceFile
{
	enum State {
		IDLE = 0,
		UNPACKING,
		RENDERING_PDF,
		CONVERTING,
		PACKING,
		DONE
	} state;

	int jobCount;

	AnsiString stateName(void)
	{
		switch (state)
		{
			case IDLE:
				if (jobCount == 0)
					return "Idle";
				else
					return "Waiting...";
			case UNPACKING:
				return "Unpacking...";
			case RENDERING_PDF:
				return "Rendering PDF...";
			case CONVERTING:
				return "Converting...";
			case PACKING:
				return "Packing...";
			case DONE:
				return "Done";
			default:
				return "???";
		}
	}

	AnsiString name;
	unsigned int size;
	unsigned int imgCount;
	unsigned int imgSize;
	ResizeCfg resizeCfg;
	bool unpacked;
	bool defaultResizeApplied;

	bool widthValid;
	AnsiString asWidth;

	bool outSizeValid;
	unsigned int outSize; 

	struct FileDesc {
		AnsiString name;
		int width;
		int height;
		FileDesc(void):
			width(-1),
			height(-1)
		{}
    };
	std::vector<FileDesc> files;
	std::map<int, int> widths;
	int mostFrequentWidth;

	SourceFile(void) {
		size = 0;
		imgCount = 0;
		imgSize = 0;
		unpacked = false;
		widthValid = false;
		outSizeValid = false;
		outSize = 0;
		defaultResizeApplied = false;
		mostFrequentWidth = 0;
		state = IDLE;
		jobCount = 0;
	}

	bool operator==(const SourceFile &b)
	{
    	return name == b.name;
	}


};

#endif
